@@ -13,21 +13,19 @@ from typing import (
     Annotated,
     Literal,
 )
-from types import (
-    FunctionType,
-    MappingProxyType,
-)
-from functools import wraps
+from types import FunctionType, MappingProxyType, new_class
+import sys
 
 from lru import LRU
 
 from ._common import (
-    MISSING,
+    Empty,
     get_annotations,
     Constraints,
     BaseConstraint,
     cached_get_origin,
     cached_get_args,
+    T_FuzzyTypeHint,
 )
 
 __all__ = [
@@ -46,10 +44,12 @@ __all__ = [
     "get_extras",
     "get_constraints",
     "struct",
+    "create_struct",
 ]
 
 T = TypeVar("T")
 P = ParamSpec("P")
+_INITIALIZED = "_initialized_"
 _POST_INIT = "_post_init_"
 _PARAMS = "_params_"
 _FIELDS = "_fields_"
@@ -66,6 +66,7 @@ _CLASS_ATTRS_DEFAULTS = {
     _CONSTRAINTS: None,
 }
 _INSTANCE_ATTRS = (
+    _INITIALIZED,
     _SETTED,
     _EXTRAS,
     _FROZEN_HASH,
@@ -90,9 +91,10 @@ class StructProto(Protocol):
     _setted_: ClassVar[set[T_FieldName]]
     _computed_: Optional[set[T_FieldName]]
     _post_init_: Optional[Callable[[Any], None]]
+    _initialized_: bool
     _constraints_: Optional[tuple["BaseConstraint"]]
     _params_: Optional[tuple[str]]
-    _extra_: Optional[dict]
+    _extras_: Optional[dict]
 
 
 class FrozenInstanceError(AttributeError):
@@ -104,8 +106,8 @@ class field:
     # a ton of setattr calls slowing down initial model creation
     name = None
     type_hint = None
-    default = MISSING
-    default_factory = MISSING
+    default = Empty
+    default_factory = Empty
     slots = False
     init = False
     repr = False
@@ -127,11 +129,16 @@ class field:
             self.is_cached = cache
 
     @property
-    def is_required(self):
+    def is_required(self) -> bool:
+        # if Required constraint explicitly provided, base off that; otherwise
+        # we infer based on other settings
+        required = self.constraints.get("required")
+        if required is not None:
+            return required.value
         return (
             not self.is_computed
-            and self.default is MISSING
-            and self.default_factory is MISSING
+            and self.default is Empty
+            and self.default_factory is Empty
         )
 
     def __repr__(self):
@@ -161,7 +168,7 @@ class field:
             setname(owner, name)
         if (computed := getattr(owner, _COMPUTED, None)) is None:
             computed = set()
-            setattr(owner, _COMPUTED, computed)
+            owner._computed_ = computed
         computed.add(name)
 
     def __get__(self, instance: Any, owner: type | None = None, /) -> Any:
@@ -189,9 +196,6 @@ class field:
         return self
 
 
-# TODO: comb through these helper functions and think through how they are used, if caching can be used, do we need them, etc.
-# decide when to use function internally vs copying code
-
 # OPTIMIZATION: skip isinstance check to determine if instance or class
 # and use separate function depending on caller context. fine for internal
 # use in hot paths since the input type is usually known
@@ -212,23 +216,17 @@ def is_struct_class(cls) -> TypeIs[StructProto]:
         return result
 
 
-_empty_fields: MappingProxyType = MappingProxyType({})
-
-
 def get_fields(obj: type[StructProto] | StructProto) -> MappingProxyType[str, "field"]:
     """Return a mapping of field name and Field object.
 
     Accepts a model or an instance of one.
     """
-    try:
-        return MappingProxyType(getattr(obj, _FIELDS))
-    except AttributeError:
-        return _empty_fields
+    return MappingProxyType(obj._fields_)
 
 
 def get_names(obj: type[StructProto] | StructProto) -> tuple[str]:
     """Return a collection of the field names"""
-    return tuple(get_fields(obj))
+    return tuple(obj._fields_.keys())
 
 
 _empty_computed = frozenset()
@@ -236,25 +234,16 @@ _empty_computed = frozenset()
 
 def get_computed(obj: type[StructProto] | StructProto) -> frozenset[str]:
     """Return a set of field names on the model which are computed"""
-    try:
-        result = getattr(obj, _COMPUTED)
-    except (
-        AttributeError,
-        TypeError,
-    ):  # not set unless there is at least one computed field on regular class
+    result = getattr(obj, _COMPUTED, None)
+    if result is None:
         return _empty_computed
-    else:
-        if result is None:  # slotted class has this set to None by default
-            return _empty_computed
-        return frozenset(result)
+    return frozenset(result)
 
 
 def get_required(obj: type[StructProto] | StructProto) -> frozenset[str]:
     """Return a set of field names on the model which are required
     (i.e. their values must be given on init)"""
-    return frozenset(
-        name for name, field in get_fields(obj).items() if field.is_required
-    )
+    return frozenset(name for name, field in obj._fields_.items() if field.is_required)
 
 
 def get_optional(obj: type[StructProto] | StructProto) -> frozenset[str]:
@@ -262,7 +251,7 @@ def get_optional(obj: type[StructProto] | StructProto) -> frozenset[str]:
     (i.e. they have default values or are computed so they dont have to be given on init)
     """
     return frozenset(
-        name for name, field in get_fields(obj).items() if not field.is_required
+        name for name, field in obj._fields_.items() if not field.is_required
     )
 
 
@@ -271,15 +260,15 @@ _default_setted = frozenset()
 
 def get_setted(obj: StructProto) -> frozenset[str]:
     """Return set of field names which were set in the model constructor."""
-    try:
-        return frozenset(getattr(obj, _SETTED))
+    try:  # setted not created unless there is at least one field and init called
+        return frozenset(obj._setted_)
     except AttributeError:
         return _default_setted
 
 
 def get_unsetted(obj: StructProto) -> frozenset[str]:
     """Return set of field names which were not set in the model constructor."""
-    return frozenset(getattr(obj, _FIELDS).keys() - get_setted(obj))
+    return frozenset(obj._fields_.keys() - get_setted(obj))
 
 
 def set_extras(obj: StructProto, extras: dict) -> None:
@@ -290,23 +279,16 @@ _extra_empty: MappingProxyType = MappingProxyType({})
 
 
 def get_extras(obj: StructProto) -> MappingProxyType:
-    try:
-        # extras not present a lot of the time so fallback to empty instead of using AttributeError
-        # to cut down on exception handling costs
-        return MappingProxyType(getattr(obj, _EXTRAS, _extra_empty))
-    except AttributeError:
-        return _extra_empty
+    # extras not present a lot of the time so fallback to empty instead of using AttributeError
+    # to cut down on exception handling costs
+    return MappingProxyType(getattr(obj, _EXTRAS, _extra_empty))
 
 
 def get_constraints(obj) -> Constraints:
-    try:
-        result = getattr(obj, _CONSTRAINTS)
-    except AttributeError:  # regular class, constraints attr not set unless defined
+    result = getattr(obj, _CONSTRAINTS, None)
+    if result is None:  # slotted class has this set to None by default
         return Constraints.empty
-    else:
-        if result is None:  # slotted class has this set to None by default
-            return Constraints.empty
-        return result
+    return result
 
 
 def __replace__(self, **changes):
@@ -321,17 +303,42 @@ def __replace__(self, **changes):
     return self.__class__(**changes)
 
 
+def __not_slotted_getitem__(self, name):
+    # ASSUMPTION: all field values are stored on the instance everytime and not stored
+    # in the class attrs or anywhere else! this must remain try or this will throw KeyError
+    # should be a bit faster than getattr and *should* be safe since field values should
+    # get stored in the instance dict
+    return self.__dict__[name]
+
+
+def __slotted_getitem__(self, name):
+    return self.__getattribute__(name)
+
+
+def __setattr__(self, name, value):
+    super(cls, self).__setattr__(name, value)  # type: ignore
+    # BUG FIX: we have to override for bookkeeping purposes. if field set to value after init
+    # we still need to make sure we add it to the set of setted fields. did not
+    # need in earlier design because structs were immutable
+    if getattr(self, _INITIALIZED, None) and name in self._fields_:
+        try:
+            self._setted_.add(name)
+        except AttributeError:
+            # avoid infinite loop
+            object.__setattr__(self, _SETTED, set((name,)))
+
+
 # cls is a global which is passed in when model defined. used to freeze setting anything
 # on that class instance, not just field names. this "resets" when it is subclassed and instead
 # only applies the constraint to the fields
 def __frozen_setattr__(self, name, value):
-    if self.__class__ is cls or name in getattr(self, _FIELDS):  # type: ignore
+    if self.__class__ is cls or name in self._fields_:  # type: ignore
         raise FrozenInstanceError(f"cannot assign to field {name!r}")
     return super(cls, self).__setattr__(name, value)  # type: ignore
 
 
 def __frozen_delattr__(self, name):
-    if self.__class__ is cls or name in getattr(self, _FIELDS):  # type: ignore
+    if self.__class__ is cls or name in self._fields_:  # type: ignore
         raise FrozenInstanceError(f"cannot delete field {name!r}")
     return super(cls, self).__delattr__(name)  # type: ignore
 
@@ -360,6 +367,7 @@ _lazy_ge = _LazyDescriptor("__ge__", "_create_comparator", ("__ge__",))
 _lazy_le = _LazyDescriptor("__le__", "_create_comparator", ("__le__",))
 _lazy_lt = _LazyDescriptor("__lt__", "_create_comparator", ("__lt__",))
 _lazy_repr = _LazyDescriptor("__repr__", "_create_repr")
+_lazy_setitem = _LazyDescriptor("__setitem__", "_create_setitem")
 _lazy_frozen_setattr = _LazyDescriptor("__setattr__", "_create_frozen")
 _lazy_frozen_delattr = _LazyDescriptor("__delattr__", "_create_frozen")
 _lazy_hash = _LazyDescriptor("__hash__", "_create_hash")
@@ -401,9 +409,13 @@ class StructGenerator:
         hash: bool,
         replace: bool,
         slots: bool,
+        getitem: bool,
+        setitem: bool,
     ) -> dict[str, field]:
         fields_dict = {}
         for k in cls.__mro__[-1:0:-1]:
+            if not is_struct_class(k):
+                continue
             if k_fields := get_fields(k):
                 fields_dict.update(k_fields)
 
@@ -427,7 +439,7 @@ class StructGenerator:
                 {
                     "name": fn,
                     "type_hint": type_hint,
-                    "default": cls_dict_get(fn, MISSING),
+                    "default": cls_dict_get(fn, Empty),
                 }
             )
             constraints = Constraints.empty
@@ -439,17 +451,17 @@ class StructGenerator:
                 ]
                 if constraints_args:
                     constraints = Constraints(constraints_args)
-                    if field_attrs.get("default", MISSING) is MISSING:
-                        if (
-                            default := constraints.get("default", MISSING)
-                        ) is not MISSING:
+                    if field_attrs.get("default", Empty) is Empty:
+                        if (default := constraints.get("default", Empty)) is not Empty:
                             field_attrs["default"] = default.value
                         if (
-                            default_factory := constraints.get(
-                                "default_factory", MISSING
-                            )
-                        ) is not MISSING:
+                            default_factory := constraints.get("default_factory", Empty)
+                        ) is not Empty:
                             field_attrs["default_factory"] = default_factory.value
+                # NOTE: do not set default value as class attr if not already set and instead
+                # always set on struct instance to speed up getting values from struct through __dict__
+                # accessing field values probably more common and done more often so likely (?)
+                # worth the extra memory per instance and overhead on init in setting attrs
             field_attrs["constraints"] = constraints
             f = field()
             f.__dict__.update(field_attrs)
@@ -486,7 +498,7 @@ class StructGenerator:
             )
             fields_dict[fn] = f
 
-        setattr(cls, _FIELDS, fields_dict)
+        cls._fields_ = fields_dict
         return fields_dict
 
     def _create_comparator(
@@ -502,6 +514,8 @@ class StructGenerator:
         hash: bool,
         replace: bool,
         slots: bool,
+        getitem: bool,
+        setitem: bool,
     ) -> None:
         field_names: tuple[str] = get_names(cls)
         exact_key = (op, field_names)
@@ -583,6 +597,8 @@ class StructGenerator:
         hash: bool,
         replace: bool,
         slots: bool,
+        getitem: bool,
+        setitem: bool,
     ) -> None:
         cls.__replace__ = __replace__
 
@@ -598,6 +614,8 @@ class StructGenerator:
         hash: bool,
         replace: bool,
         slots: bool,
+        getitem: bool,
+        setitem: bool,
     ) -> None:
         field_names: tuple[str] = get_names(cls)
         exact_key = (field_names, frozen)
@@ -696,6 +714,47 @@ class StructGenerator:
             self._cache_repr_exact[exact_key] = func
         cls.__repr__ = func
 
+    def _create_setattr(
+        self,
+        cls,
+        init: bool,
+        repr: bool,
+        eq: bool,
+        order: bool,
+        frozen: bool,
+        kw_only: bool,
+        hash: bool,
+        replace: bool,
+        slots: bool,
+        getitem: bool,
+        setitem: bool,
+    ) -> None:
+        cls.__setattr__ = __setattr__.__class__(  # type: ignore
+            __setattr__.__code__,
+            {
+                "cls": cls,
+                "_INITIALIZED": _INITIALIZED,
+                "_SETTED": _SETTED,
+            },
+        )
+
+    def _create_setitem(
+        self,
+        cls,
+        init: bool,
+        repr: bool,
+        eq: bool,
+        order: bool,
+        frozen: bool,
+        kw_only: bool,
+        hash: bool,
+        replace: bool,
+        slots: bool,
+        getitem: bool,
+        setitem: bool,
+    ) -> None:
+        cls.__setitem__ = cls.__setattr__
+
     def _create_frozen(
         self,
         cls,
@@ -708,6 +767,8 @@ class StructGenerator:
         hash: bool,
         replace: bool,
         slots: bool,
+        getitem: bool,
+        setitem: bool,
     ) -> None:
         cls.__setattr__ = __frozen_setattr__.__class__(  # type: ignore
             __frozen_setattr__.__code__,
@@ -738,6 +799,8 @@ class StructGenerator:
         hash: bool,
         replace: bool,
         slots: bool,
+        getitem: bool,
+        setitem: bool,
     ) -> None:
         field_names: tuple[str] = get_names(cls)
         exact_key = (field_names, frozen)
@@ -814,38 +877,45 @@ class StructGenerator:
             self._cache_hash_exact[exact_key] = func
         cls.__hash__ = func
 
-    def _create_init_no_fields_template(
-        self, kw_only: bool, field_count: int, has_defaults: bool
+    def _create_init_no_fields_not_frozen_template(
+        self, kw_only: bool, field_count: int, has_optional: bool
     ) -> FunctionType:
-        code_str = "def __init__(self): pass"
+        code_str = f"def __init__(self):\n    self.{_INITIALIZED} = True"
         exec(code_str, {}, l := {})
-        self._cache_init[(field_count, kw_only, has_defaults, True)] = (
-            template_function
-        ) = l.pop("__init__")
-        self._cache_init[(field_count, kw_only, has_defaults, False)] = (
+        template_function = l.pop("__init__")
+        self._cache_init[(field_count, kw_only, has_optional, False)] = (
             template_function
         )
-        self._cache_init[(field_count, not kw_only, has_defaults, True)] = (
-            template_function
-        )
-        self._cache_init[(field_count, not kw_only, has_defaults, False)] = (
+        self._cache_init[(field_count, not kw_only, has_optional, False)] = (
             template_function
         )
         return template_function
 
-    def _create_init_has_defaults_frozen_template(
-        self, kw_only: bool, field_count: int, has_defaults: bool
+    def _create_init_no_fields_frozen_template(
+        self, kw_only: bool, field_count: int, has_optional: bool
+    ) -> FunctionType:
+        code_str = f"def __init__(self):\n    obj_setattr(self, {_INITIALIZED!r}, True)"
+        exec(code_str, {"obj_setattr": object.__setattr__}, l := {})
+        template_function = l.pop("__init__")
+        self._cache_init[(field_count, kw_only, has_optional, True)] = template_function
+        self._cache_init[(field_count, not kw_only, has_optional, True)] = (
+            template_function
+        )
+        return template_function
+
+    def _create_init_has_optional_frozen_template(
+        self, kw_only: bool, field_count: int, has_optional: bool
     ) -> FunctionType:
         code_str = (
             f"def __init__(self, {'*, ' if kw_only else ''}{', '.join(f'_field_{i}' for i in range(field_count))}):\n"
             + "    setted = set()\n"
             + f'    obj_setattr(self, "{_SETTED}", setted)\n'
             + "\n".join(
-                f"    if _field_{i} is MISSING:\n"
-                f"        if _field_{i}_default is not MISSING:\n"
-                f'            obj_setattr(self, "_field_{i}", _field_{i}_default)\n'
-                f"        else:\n"
+                f"    if _field_{i} is Empty:\n"
+                f"        if _field_{i}_default_factory is not Empty:\n"
                 f'            obj_setattr(self, "_field_{i}", _field_{i}_default_factory())\n'
+                f"        else:\n"
+                f'            obj_setattr(self, "_field_{i}", _field_{i}_default)\n'
                 f"    else:\n"
                 f'        setted.add("_field_{i}")\n'
                 f'        obj_setattr(self, "_field_{i}", _field_{i})'
@@ -854,32 +924,33 @@ class StructGenerator:
             + "\n"
             + (
                 f"    if post_init := getattr(self, '{_POST_INIT}', None):\n"
-                "        post_init()"
+                "        post_init()\n"
+                f'    obj_setattr(self, "{_INITIALIZED}", True)'
             )
         )
         exec(
             code_str,
-            {"MISSING": MISSING, "obj_setattr": object.__setattr__},
+            {"Empty": Empty, "obj_setattr": object.__setattr__},
             l := {},  # type: ignore
         )
-        self._cache_init[(field_count, kw_only, has_defaults, True)] = (
+        self._cache_init[(field_count, kw_only, has_optional, True)] = (
             template_function
         ) = l.pop("__init__")
         return template_function
 
-    def _create_init_has_defaults_not_frozen_template(
-        self, kw_only: bool, field_count: int, has_defaults: bool
+    def _create_init_has_optional_not_frozen_template(
+        self, kw_only: bool, field_count: int, has_optional: bool
     ) -> FunctionType:
         code_str = (
             f"def __init__(self, {'*, ' if kw_only else ''}{', '.join(f'_field_{i}' for i in range(field_count))}):\n"
             + "    setted = set()\n"
             + f"    self.{_SETTED} = setted\n"
             + "\n".join(
-                f"    if _field_{i} is MISSING:\n"
-                f"        if _field_{i}_default is not MISSING:\n"
-                f"            self._field_{i} = _field_{i}_default\n"
-                f"        else:\n"
+                f"    if _field_{i} is Empty:\n"
+                f"        if _field_{i}_default_factory is not Empty:\n"
                 f"            self._field_{i} = _field_{i}_default_factory()\n"
+                f"        else:\n"
+                f"            self._field_{i} = _field_{i}_default\n"
                 f"    else:\n"
                 f'        setted.add("_field_{i}")\n'
                 f"        self._field_{i} = _field_{i}"
@@ -888,17 +959,22 @@ class StructGenerator:
             + "\n"
             + (
                 f"    if post_init := getattr(self, '{_POST_INIT}', None):\n"
-                "        post_init()"
+                "        post_init()\n"
+                f"    self.{_INITIALIZED} = True"
             )
         )
-        exec(code_str, {"MISSING": MISSING}, l := {})
-        self._cache_init[(field_count, kw_only, has_defaults, False)] = (
+        exec(
+            code_str,
+            {"Empty": Empty},
+            l := {},
+        )
+        self._cache_init[(field_count, kw_only, has_optional, False)] = (
             template_function
         ) = l.pop("__init__")
         return template_function
 
-    def _create_init_no_defaults_frozen_template(
-        self, kw_only: bool, field_count: int, has_defaults: bool
+    def _create_init_no_optional_frozen_template(
+        self, kw_only: bool, field_count: int, has_optional: bool
     ) -> FunctionType:
         code_str = (
             f"def __init__(self, {'*, ' if kw_only else ''}{', '.join(f'_field_{i}' for i in range(field_count))}):\n"
@@ -909,17 +985,18 @@ class StructGenerator:
             + "\n"
             + (
                 f"    if post_init := getattr(self, '{_POST_INIT}', None):\n"
-                "        post_init()"
+                "        post_init()\n"
+                f'    obj_setattr(self, "{_INITIALIZED}", True)'
             )
         )
         exec(code_str, {"obj_setattr": object.__setattr__}, l := {})
-        self._cache_init[(field_count, kw_only, has_defaults, True)] = (
+        self._cache_init[(field_count, kw_only, has_optional, True)] = (
             template_function
         ) = l.pop("__init__")
         return template_function
 
-    def _create_init_no_defaults_not_frozen_template(
-        self, kw_only: bool, field_count: int, has_defaults: bool
+    def _create_init_no_optional_not_frozen_template(
+        self, kw_only: bool, field_count: int, has_optional: bool
     ) -> FunctionType:
         code_str = (
             f"def __init__(self, {'*, ' if kw_only else ''}{', '.join(f'_field_{i}' for i in range(field_count))}):\n"
@@ -927,16 +1004,17 @@ class StructGenerator:
             + "\n"
             + (
                 f"    if post_init := getattr(self, '{_POST_INIT}', None):\n"
-                "        post_init()"
+                "        post_init()\n"
+                f"    self.{_INITIALIZED} = True"
             )
         )
         exec(code_str, {}, l := {})
-        self._cache_init[(field_count, kw_only, has_defaults, False)] = (
+        self._cache_init[(field_count, kw_only, has_optional, False)] = (
             template_function
         ) = l.pop("__init__")
         return template_function
 
-    def _fill_init_has_defaults_frozen_template(
+    def _fill_init_has_optional_frozen_template(
         self,
         init_exact_key: tuple,
         template_function: FunctionType,
@@ -944,6 +1022,7 @@ class StructGenerator:
         field_defaults: tuple,
         field_default_factories: tuple,
         kw_only: bool,
+        optional_names: tuple[str],
     ) -> FunctionType:
         func = template_function.__class__(  # type: ignore
             template_function.__code__.replace(  # type: ignore
@@ -958,6 +1037,8 @@ class StructGenerator:
                     _SETTED,
                     *field_names,
                     _POST_INIT,
+                    _INITIALIZED,
+                    True,
                 ),
             ),
             template_function.__globals__  # type: ignore
@@ -968,29 +1049,13 @@ class StructGenerator:
             ),
         )
         if kw_only:
-            func.__kwdefaults__ = (  # type: ignore
-                dict(
-                    (k, MISSING)
-                    for k, d, df in zip(
-                        field_names, field_defaults, field_default_factories
-                    )
-                    if d is not MISSING or df is not MISSING
-                )
-                or None
-            )
+            func.__kwdefaults__ = dict.fromkeys(optional_names, Empty) or None
         else:
-            func.__defaults__ = (  # type: ignore
-                tuple(
-                    MISSING
-                    for d, df in zip(field_defaults, field_default_factories)
-                    if d is not MISSING or df is not MISSING
-                )
-                or None
-            )
+            func.__defaults__ = ((Empty,) * len(optional_names)) or None
         self._cache_init_exact[init_exact_key] = func
         return func
 
-    def _fill_init_has_defaults_not_frozen_template(
+    def _fill_init_has_optional_not_frozen_template(
         self,
         init_exact_key: tuple,
         template_function: FunctionType,
@@ -998,27 +1063,29 @@ class StructGenerator:
         field_defaults: tuple,
         field_default_factories: tuple,
         kw_only: bool,
+        optional_names: tuple[str],
     ) -> FunctionType:
         func = template_function.__class__(  # type: ignore
             template_function.__code__.replace(  # type: ignore
                 co_names=(
                     "set",
                     _SETTED,
-                    "MISSING",
-                    f"_field_0_default",
-                    field_names[0],
+                    "Empty",
                     f"_field_0_default_factory",
+                    field_names[0],
+                    f"_field_0_default",
                     "add",
                     *tuple(
                         con
                         for i, field_name in enumerate(field_names[1:])
                         for con in (
-                            f"_field_{i+1}_default",
-                            field_name,
                             f"_field_{i+1}_default_factory",
+                            field_name,
+                            f"_field_{i+1}_default",
                         )
                     ),
                     "getattr",
+                    _INITIALIZED,
                 ),
                 co_varnames=(
                     "self",
@@ -1026,7 +1093,7 @@ class StructGenerator:
                     "setted",
                     "post_init",
                 ),
-                co_consts=(None, *field_names, _POST_INIT),
+                co_consts=(None, *field_names, _POST_INIT, True),
             ),
             template_function.__globals__  # type: ignore
             | dict((f"_field_{i}_default", d) for i, d in enumerate(field_defaults))
@@ -1036,25 +1103,9 @@ class StructGenerator:
             ),
         )
         if kw_only:
-            func.__kwdefaults__ = (  # type: ignore
-                dict(
-                    (k, MISSING)
-                    for k, d, df in zip(
-                        field_names, field_defaults, field_default_factories
-                    )
-                    if d is not MISSING or df is not MISSING
-                )
-                or None
-            )
+            func.__kwdefaults__ = dict.fromkeys(optional_names, Empty) or None
         else:
-            func.__defaults__ = (  # type: ignore
-                tuple(
-                    MISSING
-                    for d, df in zip(field_defaults, field_default_factories)
-                    if d is not MISSING or df is not MISSING
-                )
-                or None
-            )
+            func.__defaults__ = ((Empty,) * len(optional_names)) or None
         self._cache_init_exact[init_exact_key] = func
         return func
 
@@ -1071,7 +1122,7 @@ class StructGenerator:
                     *field_names,
                     "post_init",
                 ),
-                co_consts=(None, *field_names, _POST_INIT),
+                co_consts=(None, *field_names, _POST_INIT, _INITIALIZED, True),
             ),
             template_function.__globals__,
         )
@@ -1089,6 +1140,7 @@ class StructGenerator:
                 co_names=(
                     *field_names,
                     "getattr",
+                    _INITIALIZED,
                 ),
                 co_varnames=("self", *field_names, "post_init"),
             ),
@@ -1109,6 +1161,8 @@ class StructGenerator:
         hash: bool,
         replace: bool,
         slots: bool,
+        getitem: bool,
+        setitem: bool,
     ) -> None:
         fields_dict = get_fields(cls)
         field_names: tuple[str] = tuple(
@@ -1134,89 +1188,99 @@ class StructGenerator:
             # to use kwargs or something in init so this is not an issue, but should generally just
             # stick to keyword only when possible anyway
             if not kw_only:
-                defaulted = None
+                first_optional_name = None
                 for f in fields_dict.values():
-                    if f.is_computed:
-                        continue
-                    if f.default is not MISSING or f.default_factory is not MISSING:
-                        defaulted = f.name
-                    elif defaulted:
+                    if not f.is_required:
+                        first_optional_name = f.name
+                    elif first_optional_name:
                         raise TypeError(
-                            f"non-default argument, {f.name!r}, "
-                            f"follows default argument, {defaulted!r}, in model, {cls!r}"
+                            f"required field, {f.name!r}, "
+                            f"after optional field, {first_optional_name!r}, in model, {cls!r}"
                         )
-            has_defaults = any(_ is not MISSING for _ in field_defaults) or any(
-                _ is not MISSING for _ in field_default_factories
+            optional_names = tuple(
+                fn for fn in field_names if not fields_dict[fn].is_required
             )
+            has_optional = bool(optional_names)
             field_count = len(field_names)
             try:
                 template_function = self._cache_init[
-                    (field_count, kw_only, has_defaults, frozen)
+                    (field_count, kw_only, has_optional, frozen)
                 ]
             except KeyError:
                 # special case: cannot use * for kw only if nothing follows it
                 if field_count == 0:
-                    template_function = self._create_init_no_fields_template(
-                        kw_only=kw_only,
-                        field_count=field_count,
-                        has_defaults=has_defaults,
-                    )
-                elif has_defaults:
                     if frozen:
+                        template_function = self._create_init_no_fields_frozen_template(
+                            kw_only=kw_only,
+                            field_count=field_count,
+                            has_optional=has_optional,
+                        )
+                    else:
                         template_function = (
-                            self._create_init_has_defaults_frozen_template(
+                            self._create_init_no_fields_not_frozen_template(
                                 kw_only=kw_only,
                                 field_count=field_count,
-                                has_defaults=has_defaults,
+                                has_optional=has_optional,
+                            )
+                        )
+                elif has_optional:
+                    if frozen:
+                        template_function = (
+                            self._create_init_has_optional_frozen_template(
+                                kw_only=kw_only,
+                                field_count=field_count,
+                                has_optional=has_optional,
                             )
                         )
                     else:
                         template_function = (
-                            self._create_init_has_defaults_not_frozen_template(
+                            self._create_init_has_optional_not_frozen_template(
                                 kw_only=kw_only,
                                 field_count=field_count,
-                                has_defaults=has_defaults,
+                                has_optional=has_optional,
                             )
                         )
                 else:
                     if frozen:
                         template_function = (
-                            self._create_init_no_defaults_frozen_template(
+                            self._create_init_no_optional_frozen_template(
                                 kw_only=kw_only,
                                 field_count=field_count,
-                                has_defaults=has_defaults,
+                                has_optional=has_optional,
                             )
                         )
                     else:
                         template_function = (
-                            self._create_init_no_defaults_not_frozen_template(
+                            self._create_init_no_optional_not_frozen_template(
                                 kw_only=kw_only,
                                 field_count=field_count,
-                                has_defaults=has_defaults,
+                                has_optional=has_optional,
                             )
                         )
             # NOTE: should maybe technically replace _field_{x}_default_factory with field_name_default_factory
             # but this works as-is and no one will see it so skipping
             if field_count == 0:
                 func = template_function
-            elif has_defaults:
+            elif has_optional:
                 if frozen:
-                    func = self._fill_init_has_defaults_frozen_template(
+                    func = self._fill_init_has_optional_frozen_template(
                         init_exact_key=init_exact_key,
                         template_function=template_function,
                         field_names=field_names,
                         field_defaults=field_defaults,
                         field_default_factories=field_default_factories,
                         kw_only=kw_only,
+                        optional_names=optional_names,
                     )
                 else:
-                    func = self._fill_init_has_defaults_not_frozen_template(
+                    func = self._fill_init_has_optional_not_frozen_template(
                         init_exact_key=init_exact_key,
                         template_function=template_function,
                         field_names=field_names,
                         field_defaults=field_defaults,
                         field_default_factories=field_default_factories,
                         kw_only=kw_only,
+                        optional_names=optional_names,
                     )
             else:
                 if frozen:
@@ -1305,19 +1369,33 @@ class StructGenerator:
         hash: bool,
         replace: bool,
         slots: bool,
+        getitem: bool,
+        setitem: bool,
         constraints: Iterable["BaseConstraint"],
     ) -> type[T]:
         # NOTE: lazy building model to minimize import costs. separate fields from methods
         # since during building of codec we only need the fields and can skip building the methods
         # which are a bit slower due to dynamic code generation
         if constraints:
-            setattr(cls, _CONSTRAINTS, Constraints(constraints))
-        setattr(
-            cls,
-            _PARAMS,
-            (self, (init, repr, eq, order, frozen, kw_only, hash, replace, slots)),
+            cls._constraints_ = Constraints(constraints)
+        args = (
+                init,
+                repr,
+                eq,
+                order,
+                frozen,
+                kw_only,
+                hash,
+                replace,
+                slots,
+                getitem,
+                setitem,
+            )
+        cls._params_ = (
+            self,
+            args
         )
-        setattr(cls, _FIELDS, _lazy_fields)
+        cls._fields_ = _lazy_fields
         if slots:
             # __slots__ must be defined when the class is created or it has no effect
             # using a proxy to do lazy loading would defeat perf gains and create problems when importing
@@ -1337,10 +1415,20 @@ class StructGenerator:
         if frozen:
             cls.__setattr__ = _lazy_frozen_setattr
             cls.__delattr__ = _lazy_frozen_delattr
+        else:
+            # BUG FIX: materialize __setattr__ immediately when mutable to avoid infinite recursion 
+            self._create_setattr(cls, *args)
         if hash:
             cls.__hash__ = _lazy_hash
         if replace:
             cls.__replace__ = _lazy_replace
+        if getitem:
+            if slots:
+                cls.__getitem__ = __slotted_getitem__
+            else:
+                cls.__getitem__ = __not_slotted_getitem__
+        if setitem:
+            cls.__setitem__ = _lazy_setitem
         return cls
 
 
@@ -1353,7 +1441,7 @@ _model_gen = StructGenerator()
     eq_default=True,
     order_default=True,
     kw_only_default=True,
-    frozen_default=True,
+    frozen_default=False,
     field_specifiers=(field,),
 )
 def struct(
@@ -1369,6 +1457,8 @@ def struct(
     hash: bool = True,
     replace: bool = True,
     slots: bool = False,  # defaults to off due to significant overhead added in building class
+    getitem: bool = False,
+    setitem: bool = False,
     constraints: Iterable["BaseConstraint"] = (),
 ) -> type[T] | Callable[[type[T]], type[T]]:
     def wrap_model(cls):
@@ -1383,9 +1473,70 @@ def struct(
             hash=hash,
             replace=replace,
             slots=slots,
+            getitem=getitem,
+            setitem=setitem,
             constraints=constraints,
         )
 
     if cls:
         return wrap_model(cls)
     return wrap_model
+
+
+def create_struct(
+    name: str,
+    fields: dict[str, T_FuzzyTypeHint | field],
+    /,
+    *,
+    bases: tuple[type, ...] = (),
+    namespace: dict | None = None,
+    module: str | None = None,
+    init: bool = True,
+    repr: bool = True,
+    eq: bool = True,
+    order: bool = True,
+    frozen: bool = False,  # perf: faster setting of attrs on init if False
+    kw_only: bool = True,
+    hash: bool = True,
+    replace: bool = True,
+    slots: bool = False,  # defaults to off due to significant overhead added in building class
+    getitem: bool = True,  # defaults to True because that is usually main way fields worked with for these structs
+    setitem: bool = True,  # defaults to True because that is usually main way fields worked with for these structs
+    constraints: Iterable["BaseConstraint"] = (),
+) -> type[StructProto]:
+    def _setup(cls_ns: dict):
+        cls_ns.update(namespace or {})
+        cls_ns["__annotations__"] = {
+            field_name: value
+            for field_name, value in fields.items()
+            if value.__class__ is not field
+        }
+
+    cls = new_class(name=name, bases=bases, exec_body=_setup)
+
+    if module is None:
+        try:
+            module = sys._getframemodulename(1) or "__main__"
+        except AttributeError:
+            try:
+                module = sys._getframe(1).f_globals.get("__name__", "__main__")
+            except (AttributeError, ValueError):
+                pass
+    if module is not None:
+        cls.__module__ = module
+
+    return struct(
+        cls,
+        init=init,
+        repr=repr,
+        eq=eq,
+        order=order,
+        frozen=frozen,
+        kw_only=kw_only,
+        hash=hash,
+        replace=replace,
+        slots=slots,
+        getitem=getitem,
+        setitem=setitem,
+        constraints=constraints,
+    )
